from typing import Dict, Union, Any

import numpy as np
import onnxruntime
import onnx

from .abs import Basemodel
from ..exceptions import NotLoadableONNX


def create_outputs(outputs:Dict[Union[str, int], Any], results:Dict[Union[str, int], Any]):
    if len(outputs) != len(results):
        raise ValueError("The length of 'outputs' and 'results' must be the same.")

    for (key, value), result in zip(outputs.items(), results):
        outputs[key] = result
    
    return outputs


class ONNX(Basemodel):
    def __init__(self, model_file_path:str, num_threads=1):
        try:
            self.model_obj = onnxruntime.InferenceSession(model_file_path)
        except:
            raise NotLoadableONNX()
        self.inputs, self.outputs = self.model_input_output_attributes(model_file_path)

    def model_input_output_attributes(self, model_file_path:str):
        inputs = {}
        outputs = {}
        try:
            model = onnx.load(model_file_path)
        except:
            raise NotLoadableONNX()

        for inp in model.graph.input:
            shape = str(inp.type.tensor_type.shape.dim)
            inputs[inp.name] = [int(s) for s in shape.split() if s.isdigit()]

        for oup in model.graph.output:
            shape = str(oup.type.tensor_type.shape.dim)
            outputs[oup.name] = [int(s) for s in shape.split() if s.isdigit()]
        
        return inputs, outputs

    def inference(self, preprocess_result: Dict[int, np.ndarray]) -> Dict[int, np.ndarray]:
        inputs = {}
        for k, v in preprocess_result.items():
            ortvalue = onnxruntime.OrtValue.ortvalue_from_numpy(preprocess_result[k])
            inputs[k] = ortvalue
            # TODO: make function which return npy generator when len(preprocess_result[k]) > 1
        
        outputs_list = []
        for k in self.outputs.keys():
            outputs_list.append(k)

        results = self.model_obj.run(outputs_list, inputs)
        output_dict = create_outputs(self.outputs, results)
        return output_dict


import os
from typing import Dict, List, Tuple, Union

import numpy as np

from .abs import Basemodel
from ..enums import EnumInputNodeShapeFormat, EnumNodeRawDataType
from ..exceptions import NotLoadableTFLITE

try:
    import tflite_runtime.interpreter as tflite
except ImportError:
    try:
        import tensorflow as tf

        tflite = tf.lite
    except ImportError:
        raise ImportError("Failed to load tensorflow")


class DataAttribute:
    def __init__(self):
        self.attributes = self.get_props()
        for i in self.attributes:
            setattr(self, f"_{i}", None)

    def __dir__(self):
        return self.attributes

    def __iter__(self):
        for i in dir(self):
            yield i, getattr(self, i)

    def __repr__(self):
        return f"DataAttribute class of '{'location' if self.location is not None else 'name'} {self.key}' layer"

    @property
    def key(self) -> Union[int, str]:
        return self.location if self.location is not None else self.name

    @property
    def shape(self) -> Union[None, Tuple]:
        return self._shape

    @shape.setter
    def shape(self, value: Tuple):
        if not isinstance(value, tuple):
            msg = f"{__class__}.shape should be tuple, but got {type(value)}."
            raise ValueError(msg)
        self._shape = value

    @property
    def dtype(self) -> Union[None, str]:
        return self._dtype

    @dtype.setter
    def dtype(self, value):
        self._dtype = EnumNodeRawDataType(value).value

    @property
    def quantization(self) -> Union[None, tuple]:
        return self._quantization

    @quantization.setter
    def quantization(self, value):
        self._quantization = value

    @property
    def format(self) -> Union[None, str]:
        return self._format

    @format.setter
    def format(self, value):
        self._format = EnumInputNodeShapeFormat(value).value

    @property
    def location(self) -> Union[None, int]:
        """Union[None, int]: location of the Node"""
        return self._location

    @location.setter
    def location(self, value):
        self._location = value

    @property
    def name(self) -> Union[None, str]:
        """Union[None, str]: name of the Node"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def height(self) -> Union[None, int]:
        """Union[None, int]: returns height of the shape."""
        if self._height is None:
            if (self._shape and self._format) is None:
                return None
            if len(self._format) != len(self._shape):
                return None
            self._height = dict(zip(self._format, self._shape)).get("h")
        return self._height

    @property
    def width(self) -> Union[None, int]:
        """Union[None, int]: returns width of the shape."""
        if self._width is None:
            if (self._shape and self._format) is None:
                return None
            if len(self._format) != len(self._shape):
                return None
            self._width = dict(zip(self._format, self._shape)).get("w")
        return self._width

    @classmethod
    def get_props(cls) -> List:
        """returns all class properties as List."""
        return [x for x in dir(cls) if isinstance(getattr(cls, x), property)]


class TFLITE(Basemodel):
    def __init__(self, model_file_path:str, num_threads:int=1):
        try:
            with open(model_file_path, "rb") as f:
                self.interpreter_obj = tflite.Interpreter(model_content=f.read(), num_threads=num_threads)
        except:
            raise NotLoadableTFLITE()
        self.interpreter_obj.allocate_tensors()
        self.inputs, self.outputs = self.model_input_output_attributes()

    def model_input_output_attributes(self):
        inputs = {}
        outputs = {}

        for input_detail in self.interpreter_obj.get_input_details():
            input_data_attribute = DataAttribute()
            input_data_attribute.name = input_detail.get("name")
            input_data_attribute.location = input_detail.get("index")
            input_data_attribute.shape = tuple(input_detail.get("shape"))
            input_data_attribute.dtype = input_detail.get("dtype").__name__
            input_data_attribute.quantization = input_detail.get("quantization")
            inputs[input_data_attribute.key] = input_data_attribute

        for output_detail in self.interpreter_obj.get_output_details():
            output_data_attribute = DataAttribute()
            output_data_attribute.name = output_detail.get("name")
            output_data_attribute.location = output_detail.get("index")
            output_data_attribute.shape = tuple(output_detail.get("shape"))
            output_data_attribute.dtype = output_detail.get("dtype").__name__
            output_data_attribute.quantization = output_detail.get("quantization")
            outputs[output_data_attribute.key] = output_data_attribute

        return inputs, outputs    

    def inference(self, preprocess_result: Dict[int, np.ndarray], **kwargs) -> Dict[int, np.ndarray]:
        for k, v in self.inputs.items():
            if v.dtype in [np.uint8, np.int8, "int8", "unit8"]:
                pass

        for location, value in iter(preprocess_result.items()):
            self.interpreter_obj.set_tensor(location, value)
            # TODO: make function which return npy generator when len(preprocess_result[k]) > 1
        self.interpreter_obj.invoke()

        output_dict = {}
        for output_location in iter(self.outputs):
            output_dict[output_location] = self.interpreter_obj.get_tensor(output_location)
        return output_dict


import abc
from typing import Any
import os


class Basemodel(metaclass=abc.ABCMeta):    
    @abc.abstractmethod
    def model_input_output_attributes(self):
        raise NotImplementedError()
    
    @abc.abstractmethod
    def inference(self, **kwargs):
        raise NotImplementedError()


import os
from pathlib import Path
from typing import Dict, Any, Union

import numpy as np
from loguru import logger

from ..enums import Enums
from ..utils import extract_archive
from ..exceptions import NumpyLoadError, WrongDatsetFile


class NumpyDataLoader:
    def __init__(self, dataset_file_path:str, inputs:Dict[Union[str, int], Any]):
        self.npy = self.load_datasets(dataset_file_path, inputs)
    
    def load_datasets(
        self, 
        dataset_file_path:str, 
        inputs:Dict[Union[str, int], Any]
        )->Dict[Union[str, int], np.ndarray]:
        outputs = {}
        # for single numpy file case
        if Path(dataset_file_path).is_file() and Path(dataset_file_path).suffix == ".npy":
            for key in inputs.keys():
                try:
                    outputs[key]=np.load(dataset_file_path)
                except ValueError:
                    raise NumpyLoadError(detail=f"Failed to load the .npy file: {dataset_file_path}.")
            return outputs
        # for multiple numpy file case
        elif Path(dataset_file_path).suffix not in Enums.NUMPY_FILE_SUFFIXES.value:
            # Unzip dataset_file_path
            extract_archive(dataset_file_path, os.path.dirname(dataset_file_path))
            parent_directory = os.path.dirname(dataset_file_path)
            for key in inputs.keys():
                file_name = f"{key}.npy"
                file_path = os.path.join(parent_directory, file_name)
                if os.path.exists(file_path):
                    try:
                        outputs[key] = np.load(file_path)
                    except ValueError:
                        raise NumpyLoadError(detail=f"Failed to load the .npy file: {file_path}.")
                else:
                    logger.warning(f"Warning: File {file_path} does not exist.")
            return outputs
        else:
            raise WrongDatsetFile()


from pathlib import Path
from setuptools import setup, find_packages

install_requires = Path("requirements.txt").read_text().split('\n')

setup(
    name="netspresso_inference_package",
    version="0.1.5",
    author="NetsPresso",
    author_email="netspresso@nota.ai",
    description="Inference module.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/nota-github/netspresso_inference_package",
    install_requires=install_requires,
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



from enum import Enum

class Enums(Enum):
    FRAMEWORK = "framework"
    INPUTS = "inputs"
    OUTPUTS = "outputs"
    ONNX = "onnx"
    TFLITE = "tflite"
    NUMPY_FILE_SUFFIXES = [".npy", ".npz"]


class EnumInputNodeShapeFormat(Enum):
    NCHW = "nchw"
    NCWH = "ncwh"
    NHWC = "nhwc"
    NWHC = "nwhc"

    CHW = NCHW
    CWH = NCWH
    HWC = NHWC
    WHC = NWHC

    # tensorrt format
    LINEAR = NCHW
    CHW2 = NCHW
    HWC8 = NCHW
    CHW4 = NCHW
    CHW16 = NCHW
    CHW32 = NCHW

    UNKNOWN = "unknown"

    def __str__(self) -> str:
        return self.value

    @classmethod
    def _missing_(cls, value):
        value = str(value).upper()
        try:
            return cls[value]
        except KeyError:
            msg = f"{cls.__name__} expected {', '.join(list(cls.__members__.keys()))} but got `{value}`"
            raise KeyError(msg)


class EnumNodeRawDataType(Enum):
    # numpy data type
    FLOAT16 = "float16"
    FLOAT32 = "float32"
    FLOAT64 = "float64"
    INT8 = "int8"
    INT16 = "int16"
    INT32 = "int32"
    INT64 = "int64"
    UINT8 = "uint8"
    UINT16 = "uint16"
    UINT32 = "utin32"
    UINT64 = "uint64"
    # BOOL='bool_'

    # openvino data type
    FP16 = FLOAT16
    FP32 = FLOAT32
    FP64 = FLOAT64
    I8 = INT8
    I16 = INT16
    I32 = INT32
    I64 = INT64
    U8 = UINT8
    U16 = UINT16
    U32 = UINT32
    U64 = UINT64

    # tensorrt data type
    FLOAT = FLOAT32
    HALF = FLOAT16

    UNKNOWN = "unknown"

    def __str__(self) -> str:
        return self.value

    @classmethod
    def _missing_(cls, value):
        value = str(value).upper()
        try:
            return cls[value]
        except KeyError:
            msg = f"{cls.__name__} expected {', '.join(list(cls.__members__.keys()))} but got `{value}`"
            raise KeyError(msg)